_ENV = AIContext

-- code in this file was copied from pop_growth.lua because it was local there ;(
-- or otherwise could not be accessed without ripping out piece of code elsewhere

local rate2i = {
  VeryHigh=5,
  High=4,
  Medium = 3,
  Low = 2,
  VeryLow = 1,
  None=0,
}
local i2rate = {}
for rate,i ; rate2i
  i2rate[i]=rate
end

local rate2m = {
  VeryHigh=.035,
  High=.025,
  Medium = .018,
  Low = .01,
  VeryLow = .008,
  None=0.0,
}

function city_planning_growth_increase(planet)
  local imax=rate2i.VeryHigh
  local imin=rate2i.VeryLow

  local race2growth = planet:growth()
  local growth_rates = planet:growth_rates()

  local extra_growth = {}
  for race, growth ; race2growth
    local rate = growth_rates[race]
    if rate and not RACE_INFO[race].sterile
      local idefault = rate2i[rate]
      local icityplan = clamp(rate2i[rate] + 1, imin, imax)

      local mdefault = rate2m[i2rate[idefault]]
      local mcityplan = rate2m[i2rate[icityplan]]

      local extra_mult = 1.5 * mcityplan / mdefault
      --print(..race, ..growth, ..extra_mult)
      extra_growth[race] = growth * (extra_mult - 1)
    else
      extra_growth[race] = 0
    end
  end

  return extra_growth
end

local planet_field_to_race_field = {
  factories = 'hammer_production',
  labs = 'beaker_production',
  markets = 'coin_production',
  --racial food production seems to be handled by the food curve, so disregard it here
}

local accumulates_resources = set('labs', 'markets', 'mines')

function racial_production(race, field)
  if planet_field_to_race_field[field]
    local race_field = planet_field_to_race_field[field]
    return RACE_INFO[race][race_field]
  end
  return 0
end

--returns true if increasing population growth this turn will yield more resource
--than building the specified improvement. extra_growth should be a table mapping
-- race => population increase due to city planning/manufacture pop activity
local function pop_growth_production(planet, improvement, extra_growth)
  local field = planet_improvement_to_field[improvement]
  local _TRACE_INFO=field

  if field == 'machine_altars'
    field = 'factories'
  end

  local staffing_curve = planet:staffing_curves()[field]
  if not staffing_curve
    return false
  end

  --compare building the improvement now vs city planning for 1 turn then building the improvement
  --if we city plan for 1 turn, the increased production next turn accumulates resources for the 
  --entire time the improvement is under construction. however, the improvement takes one extra turn
  --to build. 

  --So for example, if the improvement takes N=3 turns to build, increases production by P once built,
  --and if city planning increases production by C due to staffing and pop growth:
  --
  --TURN #     1 2 3 4 5
  --build now: B B B P P
  --city plan: C B B B P
  --
  --since turn 5 is the same for both we can ignore it. So building now yields P/(N+1) and
  --city planning yields C*N
  --
  --HOWEVER, this logic only applies to resources that accumulate

  local build_time = estimate_improvement_build_time(planet, improvement)

  -- production increase from not city planning
  local build_now_prod = staffing_curve.base_per/(build_time + 1)

  -- city planning production increase due to staffing
  local planet_pop = pop_fun(planet)

  local staffing_prod = 0
  if staffing_curve and planet_pop < staffing_curve.staff_per*planet[field]
    local extra_pop_growth = sum(extra_growth)
    staffing_prod = extra_pop_growth*staffing_curve.active_per/staffing_curve.staff_per
  end

  -- city planning production increase due to extra pop production

  local pop_growth_prod = 0
  for race, growth ; extra_growth
    --print(..race, ..field, racial_production(race, field))
    pop_growth_prod += growth * racial_production(race, field)
  end

  if accumulates_resources[field]
    return (staffing_prod + pop_growth_prod)*build_time > build_now_prod
  end
  return (staffing_prod + pop_growth_prod) > build_now_prod
end

--ok, so suppose we want to build an improvement.
--but maybe we can do better by city planning? (e.g. due to staffing levels)
function possibly_set_city_planning(planet, improvement)
  if planet:cant_set_activity('City Planning') or planet.empire:cant_use_activity('City Planning')
    return false
  end

  local city_planning_growth = city_planning_growth_increase(planet)
  return pop_growth_production(planet, improvement, city_planning_growth)
end

function possibly_manufacture_population(planet, improvement)
  if planet:cant_set_activity('Manufacture Population') or planet.empire:cant_use_activity('Manufacture Population')
    return false
  end

  local manufactured_growth = planet:manufactured_pop_growth() --much nicer interface than for city planning :D
  if not manufactured_growth
    return false
  end

  return pop_growth_production(planet, improvement, manufactured_growth)
end
