depends "city_planning_dispatch"

_ENV = AIContext


Planet.staffing_curves = weak_memoizeNN | function(planet)
  return {
    factories = factory_curves![planet.factory_level],
    machine_altars = factory_curves![planet.factory_level],
    labs = lab_curves![planet.lab_level],
    markets = market_curves![planet.market_level],
    farms = planet.food_curve,
    mines = planet.mine_curve,
  }
end

function estimate_improvement_build_time(planet, improvement)
  if planet.raw_hammer_production == 0
    return 10000
  end

  local hammer_cost = planet:improvement_cost(improvement)

  --consider any pre-existing progress
  local hammers = 0
  if planet.building_improvement == improvement
    hammers = planet.hammers
  else
    for i,v ; planet.incomplete_projects?
      if v.building_improvement==improvement
        --use this prior project
        hammers = v.hammers
        break
      end
    end
  end

  hammer_cost = max(hammer_cost - hammers, 0)
  if hammer_cost == 0
    return 0
  end

  --consider hammers
  return max(hammer_cost - planet.overflow_hammers, 0)/planet.raw_hammer_production
end



function Planet.suggested_improvement_spread(planet)
  local prefs = setmetatable({}, sparse_metatable)

  local metal_income = empire.total_metal_production

  if planet.minerals == 'Rich'
    prefs.mines=2.2
  elseif planet.minerals=='Normal'
    prefs.mines=1
  end

  if prefs.mines
    prefs.mines *= min(max(1.25*desired_metal_production/metal_income - 1, 0.1), 2)
  end

  local pop_slots =  planet:inhabited_max()

  if pop_slots < 10
    prefs.labs = research_desire or research_desire/2
    prefs.factories = (pop_slots/8)^2 - (prefs.mines or 0) - (prefs.labs or 0)

    if planet.fertility >= 2
      prefs.farms = 3
    else
      --don't build farms on low fertility worlds.
      --if we absolutely need more food, resource deficit dispatch will 
      --force construction of the farm for us
      prefs.farms = nil
    end

    if planet:has_special_resource 'super_grain'
      prefs.farms = 1.5*(prefs.farms or 1)
    end


  else
    prefs.labs = research_desire
    prefs.factories = 6 - (prefs.labs or 0)
  end

  if empire:known 'Market Theory'
    prefs.markets = 2 - (prefs.mines or 0) - (prefs.farms or 0)
  end

  --[[
  --don't build tons of factories if we don't have the mineral production to support it
  local hammer_production = empire.total_raw_hammer_production
  if hammer_production > sustainable_hammer_production
    prefs.factories *= (sustainable_hammer_production/hammer_production)
  end
  --]]

  if planet.empire.species == 'tinker'
    prefs.machine_altars = prefs.factories
    prefs.factories= nil
    prefs.farms = nil
  end


  local staffing_curves = planet:staffing_curves()

  --print | v2s | prefs

  for field,value ; prefs
    if value <= 0
      prefs[field] = nil
    end

    --if we couldn't possibly staff another one of this improvement, don't build it
    local staff_per, base_per, active_per in staffing_curves[field]
    local built = planet[field] or 0
    if (built + 1)*staff_per > pop_slots
      prefs[field] = (prefs[field] - 1) + base_per/(base_per + active_per)
    end
  end

  --print(planet.name, "max_improvements", planet.max_improvements)

  --print | v2s | prefs

  return normalize_table | prefs
end

--planet_field_to_improvement![field]
--planet.empty_slots
--return planet.max_improvements - planet.total_improvements

function Planet.select_prefered_improvement(planet,target)

  -- as we get closer to filling out our improvement slots, factor in build time less
  local free_slots = max(1, floor(planet.max_improvements/2))
  local impatience = max(planet.empty_slots-free_slots, 0)/max(planet.max_improvements-free_slots, 0)
  --print(planet.name, ..impatience)

  -- if we decide to build a particular improvement, how much will we actually get each turn?
  local build_fraction = {}
  for field, improvement ; planet_field_to_improvement
    local build_time = estimate_improvement_build_time(planet, improvement)
    build_fraction[field] = (1/build_time - 1)*impatience + 1
  end

  local current = {}
  for field ; planet_field_to_improvement
    current[field] = planet[field]
  end

  local potential = {}
  for field ; planet_field_to_improvement
    potential[field] = setmetatable(copy | current, sparse_metatable)
    potential[field][field] = potential[field][field] + build_fraction[field]
    potential[field].construction = 1-build_fraction[field]
    
    normalize_table(potential[field])
    potential[field].construction = nil
  end

  -- needs more generic 'cant build' hook
  if planet.cant_farm 
    potential.farms=nil
  end

  -- don't worry about defenses
  potential.defenses=nil
  normalize_table(current)

  target=target or planet : suggested_improvement_spread()

  --print(planet.name, 'target', t2s | target)

  local best= get_min_key(potential,|) function(k,v)
    local err=0
--    print(v,getmetatable | v, sparse_metatable)
    for field ; planet_field_to_improvement
      --print(field, v[field], target[field])
      local delta = abs(target[field] - v[field])
      err += delta
      --print(field,k,delta,v[field],target[field])

    end
    --print(planet.name,err,k,t2s | v )
    return err
  end

  return planet_field_to_improvement![best]

end

local function nothing_to_build(planet)
  if not planet.productive
    return true
  end

  local select_focus = {}

  if not planet:cant_set_activity 'Mining'
    select_focus.Mining = planet.metal_production*max(desired_metal_production/empire.total_metal_production-1, 0)
  end
  select_focus.Research = planet.beaker_production*research_desire
  select_focus.Trade = planet.coin_production*(empire:income() > 0 and 1 or 5)

  local best_focus = get_best_value(select_focus,|) function(a,b) return a < b end
  if best_focus and not planet:cant_set_activity(best_focus)
    order.set_activity(planet, best_focus)
    return false
  end

  return true
end

function improve_remaining_planets()

  reserve_money = 0
  local push_auto, autobuilding_planets = pusher {}

  -- for all unclaimed free planets, set them to auto producing
  for planet ; free_planets
    if ( not planet:cant_build_improvement 'Planetary Defenses' )
      -- print(planet.name,..planet.defenses,..planet.factories)
      if planet.factories > planet.defenses
      --print | planet:cant_build_improvement 'Planetary Defenses'
        order.set_building_improvement(planet, 'Planetary Defenses')
        reserve_money = 800
        goto next_planet
      end
    end
    
    if planet.can_improve and planet.productive
      local _TRACE_INFO=planet.name

      local improvement = planet:select_prefered_improvement()

      if(possibly_set_city_planning(planet, improvement))
        print(empire.name, "city planning in anticipation of improvement construction", planet.name, improvement)
        order.set_activity(planet, 'City Planning')
      elseif(possibly_manufacture_population(planet, improvement))
        print(empire.name, "manufacture pop in anticipation of improvement construction", planet.name, improvement)
        order.set_activity(planet, 'Manufacture Population')
      else
        print(empire.name, "building improvement on planet", planet.name, improvement)
        order.set_building_improvement(planet, improvement) 
      end

    elseif nothing_to_build(planet)
      push_auto | planet
    end

    ::next_planet::  
  end
 
  table.sort(autobuilding_planets,|) function(a,b)
    return lexical_compare(b.raw_coin_production, a.raw_coin_production,a._ikey,b._ikey)
  end

  for i,planet in ipairs | autobuilding_planets
     order.set_autobuild(planet)
  end  

end
